from flask import current_app, request, abort
try:
    from flask_login import current_user
except ImportError:
    current_user = None


class FlaskAccess(object):
    property_name = 'is_public_endpoint'

    def __init__(self, app=None):
        self.app = None

        if app is not None:
            self.init_app(app)

    def _check_blueprint(self, endpoint):
        blueprint_name, _ = endpoint.rsplit('.', 1)
        bp = current_app.blueprints.get(blueprint_name)
        return bool(bp and getattr(bp, self.property_name, False))

    def login_required(self):
        endpoint = request.endpoint
        config = current_app.config
        login_manager = getattr(current_app, 'login_manager', None)

        if endpoint:
            rules = [
                lambda: current_user and current_user.is_authenticated,
                lambda: login_manager and login_manager.login_view == endpoint,
                lambda: endpoint in {'static', *config['ACCESS_STATIC_ENDPOINTS']},
                lambda: getattr(current_app.view_functions[endpoint], self.property_name, False),
                lambda: self._check_blueprint(endpoint),
            ]

            if not any((rule() for rule in rules)):
                if login_manager:
                    return current_app.login_manager.unauthorized()

                abort(401)

    def init_app(self, app):
        app.config.setdefault('ACCESS_STATIC_ENDPOINTS', [])
        app.before_request(self.login_required)

    def public_required(self, f):
        setattr(f, self.property_name, True)
        return f



from __future__ import annotations
import typing as t
import re

import sqlalchemy as sa

from .session import sqla_session

if t.TYPE_CHECKING:
    from sqlalchemy import Column
    from sqlalchemy.orm import Session


IdentityArgument = t.Union[t.Any, t.Tuple[t.Any, ...], t.Dict[str, t.Any]]


__all__ = (
    'generate_slug',
    'get_primary_columns',
    'normalize_pk',
)


def generate_slug(
    slug_field: t.Any,
    slug: str,
    session: t.Optional[Session] = None,
) -> str:
    """
    Generates a unique slug based on the passed value.

    Arguments:
        slug_field: Model attribute containing slug.
        slug (str): The desired slug value.
        session (Session): SQLAlchemy session.
    """
    if session is None:
        session = sqla_session

    pattern = r'^%s(?:-([0-9]+))?$' % slug

    stmt = (
        sa.select(slug_field)
          .where(slug_field.regexp_match(pattern))
          .order_by(slug_field.desc())
          .limit(1)
    )
    found = session.scalar(stmt)

    if not found:
        return slug

    match = re.match(pattern, found)

    if match is None:
        raise AssertionError('The query found one result for the regular expression.')

    return '{}-{}'.format(slug, int(match.group(1)) + 1)


def get_primary_columns(
    model_class: t.Type[t.Any],
) -> t.Tuple[Column[t.Any], ...]:
    """Returns the primary key columns."""
    return tuple(
        c for c in sa.inspect(model_class).columns if c.primary_key
    )


def normalize_pk(
    value: IdentityArgument,
    model_class: t.Type[t.Any],
) -> t.Dict[str, t.Any]:
    """Returns the primary key with a cast as a dictionary."""
    columns = get_primary_columns(model_class)

    if not isinstance(value, tuple):
        if isinstance(value, dict):
            value = tuple(value[c.name] for c in columns)
        else:
            value = (value,)

    return {
        c.name: c.type.python_type(v) for c, v in zip(columns, value)
    }



from __future__ import annotations
import typing as t

from flask import Blueprint as _Blueprint
from flask.views import View
from flask.typing import RouteCallable

from .utils import camel_to_snake


__all__ = ('Blueprint',)


class Blueprint(_Blueprint):
    def add_url_rule(
        self,
        rule: str,
        endpoint: t.Optional[str] = None,
        view_func: t.Optional[RouteCallable] = None,
        provide_automatic_options: t.Optional[bool] = None,
        *,
        view_kwargs: t.Dict[str, t.Any] = None,
        **options: t.Any,
    ) -> None:
        if endpoint is None:
            endpoint = camel_to_snake(view_func.__name__)

        if isinstance(view_func, type):
            if not issubclass(view_func, View):
                raise ValueError('view must be a subclass of `View`.')
            view_kwargs = view_kwargs or {}
            view_func = view_func.as_view(endpoint, **view_kwargs)

        super().add_url_rule(
            rule,
            endpoint,
            view_func,
            provide_automatic_options,
            **options,
        )



from __future__ import annotations
from types import TracebackType
import typing as t

from flask import current_app
from sqlalchemy.orm import Session
from werkzeug.local import LocalProxy


__all__ = (
    'get_sqla_session',
    'sqla_session',
    'SessionMixin',
)


sqla_session = t.cast(
    Session,
    LocalProxy(lambda: get_sqla_session()),
)


def get_sqla_session() -> Session:
    """Returns the current session instance from application context."""
    ext = current_app.extensions.get('sqlalchemy')

    if ext is None:
        raise RuntimeError(
            'An extension named sqlalchemy was not found '
            'in the list of registered extensions for the current application.'
        )

    return t.cast(Session, ext.db.session)


class SessionMixin:
    """
    The mixin adds a property with the current session
    and an auto-commit context manager.
    """

    def __enter__(self) -> Session:
        return self.session

    def __exit__(
        self,
        err_type: t.Optional[t.Type[BaseException]],
        err: t.Optional[BaseException],
        traceback: t.Optional[TracebackType]
    ) -> t.Optional[bool]:
        if err is None:
            self.session.commit()
        return None

    @property
    def session(self) -> Session:
        return sqla_session



from __future__ import annotations

import re
import typing as t

from flask import (
    current_app,
    redirect,
    url_for,
    request,
    flash as _flash
)


__all__ = (
    'camel_to_list', 'camel_to_snake', 'snake_to_camel',
    'get_route_param_names', 'make_redirect', 'flash',
)


def camel_to_list(s: str, lower: bool = False) -> t.List[str]:
    """Converts a camelcase string to a list."""
    lst = re.findall(r'([A-Z][a-z0-9]+)', s) or [s]
    return [w.lower() for w in lst] if lower else lst


def camel_to_snake(name: str) -> str:
    """Converts a camelcase string to a snake case string."""
    return '_'.join(camel_to_list(name, lower=True))


def snake_to_camel(name: str) -> str:
    """Converts a snake case string to a camelcase string."""
    return ''.join(name.title().split('_'))


def get_route_param_names(endpoint):
    """
    Returns parameter names from the route.

    Arguments:
        endpoint (str): The absolute name of the endpoint.
    """
    try:
        g = current_app.url_map.iter_rules(endpoint)
        return next(g).arguments
    except KeyError:
        return {}


def make_redirect(actions, params, var_name='action'):
    """
    Depending on the button pressed in the form, it creates a 302 redirect.

    Arguments:
        actions (dict):
            Mapping action names to endpoint names.
        params (object):
            An object whose property values will be used for parameters in the URL rule.
        var_name (str):
            The name of the variable whose value is used as the name of the action.

    Example:
        from functools import partial

        post_redirect = partial(make_redirect, {
            'Save': 'update',
            'Save and Create': 'create',
            'Save and Close': 'index',
        })

        @app.route('/create')
        def create():
            # ...
            if form.validate_on_submit():
                product = Product()
                form.populate_obj(product)
                db.session.add(product)
                db.session.commit()
                return post_redirect(product)
            # ...
    """
    endpoint = actions[request.form.get(var_name)]
    return redirect(make_url(endpoint, params))


def make_url(endpoint, params, **kwargs):
    """
    Creates the URL of the specified endpoint, using the passed object as rule parameter values.

    Arguments:
        endpoint (str):
            the endpoint of the URL (name of the function).
        params (object):
            An object whose property values will be used for parameters in the URL rule.
    """
    values = {}

    for name in get_route_param_names(endpoint):
        if isinstance(params, dict):
            values[name] = params.get(name)
        else:
            values[name] = getattr(params, name, None)

    return url_for(endpoint, **values, **kwargs)


class Flash:
    """
    A wrapper for flash messages that standardizes message categories.
    """

    def __call__(self, message: str, level: str = 'info') -> None:
        _flash(message, level)

    def debug(self, message: str) -> None:
        self(message, 'debug')

    def error(self, message: str) -> None:
        self(message, 'error')

    def info(self, message: str) -> None:
        self(message, 'info')

    def success(self, message: str) -> None:
        self(message, 'success')

    def warning(self, message: str) -> None:
        self(message, 'warning')


flash = Flash()



from flask import Blueprint


bp = Blueprint('users', __name__, url_prefix='/users')


@bp.get('/')
def index():
    return [
        {'id': 1, 'firstname': 'Linus', 'lastname': 'Torvalds'},
        {'id': 2, 'firstname': 'Richard', 'lastname': 'Stallman'},
    ]



"""
The module contains functions for automating the configuration of the application object.
"""

from __future__ import annotations

import typing as t

from flask import Blueprint, Flask
from flask.cli import AppGroup
from werkzeug.utils import find_modules, import_string


__all__ = (
    'register_blueprints',
    'register_commands',
    'register_extensions',
)


AppOrBp = t.Union[Flask, Blueprint]


def get_import_prefix(app: AppOrBp) -> str:
    if app.import_name == '__main__':
        return ''
    return f'{app.import_name}.'


def get_import_path(app: AppOrBp, import_path: str) -> str:
    """Returns the absolute path to import a module or package."""
    prefix = get_import_prefix(app)
    return (prefix + import_path).strip('.')


def register_blueprints(
    app: AppOrBp,
    import_path: str,
    recursive: bool = False,
    include_packages: bool = False,
) -> None:
    """
    Registers Blueprint for the specified application.

    The argument `import_path` must be a valid import name for the package that contains the modules.
    One module - one Blueprint.
    The variable named `bp` must contain an instance of Blueprint.

    If the `BLUEPRINT_DISABLED` attribute is set in the module, then Blueprint will be ignored.
    """
    modules_names = list(find_modules(
        get_import_path(app, import_path),
        recursive=recursive,
        include_packages=include_packages,
    ))

    for name in modules_names:
        mod = import_string(name)

        if hasattr(mod, 'bp') and not getattr(mod.bp, 'BLUEPRINT_DISABLED', False):
            if isinstance(mod.bp, Blueprint):
                app.register_blueprint(mod.bp)


def register_commands(app: Flask, import_name: str) -> None:
    """Initializes console commands found at the specified import path.

    If the __all__ attribute is specified in the module,
    it will be used to fund commands.
    Otherwise, the search is performed using the `dir` function.

    Command is an object inherited from `flask.cli.AppGroup`.
    """
    m = import_string(get_import_path(app, import_name))

    for name in getattr(m, '__all__', dir(m)):
        prop = getattr(m, name)

        if isinstance(prop, AppGroup):
            app.cli.add_command(prop)


def register_extensions(app: Flask, import_name: str) -> None:
    """Initializes all Flask extensions found in the specified import path.

    If the __all__ attribute is specified in the module,
    it will be used to search for extension instances.
    Otherwise, the search is performed using the `dir` function.

    An extension is an object that has an init_app method.
    """
    m = import_string(get_import_path(app, import_name))

    for name in getattr(m, '__all__', dir(m)):
        prop = getattr(m, name)
        init_app = getattr(prop, 'init_app', None)

        if callable(init_app):
            init_app(app)



from __future__ import annotations
import typing as t

from marshmallow.validate import (
    Validator,
    ValidationError,
)
from sqlalchemy import inspect, select

from .session import sqla_session
from .utils import get_primary_columns


__all__ = (
    'ExistsEntity',
    'UniqueEntity',
)


class ExistsEntity(Validator):
    """The validator checks that an entity exists."""

    default_error_messages = {
        'invalid': 'An instance with %(name)s=%(value)s does not exist.',
        'plural': 'An instance with %(attrs)s attributes does not exist.',
    }

    def __init__(
        self,
        model_class: t.Type[t.Any],
        columns: t.Optional[t.Sequence[str]] = None,
        error: str = '',
    ) -> None:
        """
        Arguments:
            model_class:
                Reference to the model class.
            columns (iterable):
                Model attributes that must be unique.
            error (str):
                Error message.
        """
        if columns is None:
            self.columns = get_primary_columns(model_class)
        else:
            ins = inspect(model_class).columns
            self.columns = tuple(ins[c] for c in columns)

        self.error = error

    def __call__(self, *values: t.Any) -> None:
        if not self._validate(*values):
            if len(values) == 1:
                raise self.make_error(
                    'invalid',
                    name=self.columns[0].name,
                    value=values[0],
                )
            else:
                raise self.make_error(
                    'plural',
                    attrs=', '.join(
                        f'{c.name}={v}' for c, v in zip(self.columns, values)
                    )
                )

    def _validate(self, *values: t.Any) -> bool:
        criteria = tuple(c == v for c, v in zip(self.columns, values))
        return bool(sqla_session.scalar(
            select(1).where(*criteria)
        ))

    def make_error(self, key: str, **kwargs: str) -> ValidationError:
        message = self.error or self.default_error_messages[key]
        return ValidationError(message % kwargs)


class UniqueEntity(ExistsEntity):
    """
    The validator checks the attributes of an entity for uniqueness.
    """

    default_error_messages = {
        'invalid': 'An instance with %(name)s=%(value)s already exists.',
        'plural': 'An instance with unique %(attrs)s attributes already exists.',
    }

    def _validate(self, *values: t.Any) -> bool:
        return not super()._validate(*values)



from wtforms import fields
from wtforms import widgets
from wtforms.compat import text_type
from wtforms.validators import StopValidation

from .widgets import CheckboxGroup


__all__ = (
    'CheckboxGroupField', 'RadioBoxField',
    'ValueListField',
)


class CheckboxGroupField(fields.SelectMultipleField):
    """
    A multiple-select, except displays a list of checkboxes.

    Iterating the field will produce subfields, allowing custom rendering of
    the enclosed checkbox fields.
    """
    widget = CheckboxGroup(inline=True)
    option_widget = widgets.CheckboxInput()


class RadioBoxField(fields.SelectField):
    """
    A multiple-select, except displays a list of checkboxes.

    Iterating the field will produce subfields, allowing custom rendering of
    the enclosed checkbox fields.
    """
    widget = CheckboxGroup(inline=True)
    option_widget = widgets.RadioInput()


class ValueListField(fields.StringField):
    """
    List of values.

    For example, can be used to enumerate tags or scope for a REST API client.

    Parameters:
        label (str): The label of the field.
        separator (str): Separator character for values.
        coerce (callable): Callback to explicitly cast each value to the desired data type.
    """
    def __init__(self, label=None, separator=',', coerce=text_type, **kwargs):
        super().__init__(label, **kwargs)
        self.separator = separator
        self.coerce = coerce
        self.data = ()

    def _value(self):
        return self.separator.join(str(v) for v in self.data)

    def process_data(self, value):
        try:
            self.data = tuple(self.coerce(v) for v in value)
        except (ValueError, TypeError):
            self.data = ()

    def process_formdata(self, valuelist):
        if valuelist and valuelist[0]:
            self.data = tuple(self.coerce(v.strip()) for v in valuelist[0].split(self.separator))



import re

import phonenumbers
from wtforms.validators import ValidationError, StopValidation


class OnlyOne(object):
    """
    Validates that a value was entered in only one of the specified form fields.
    """

    def __init__(self, fieldnames, message=None):
        self.fieldnames = fieldnames.split() if isinstance(fieldnames, str) else fieldnames
        self.message = message

    def __call__(self, form, field):
        fieldnames = [field.short_name, *self.fieldnames]

        try:
            values = [bool(form[i].data) for i in fieldnames]
        except KeyError as err:
            field_name, *_ = err.args
            raise StopValidation(field.gettext("Invalid field name '%s'.") % field_name)

        only_one = values.count(True) == 1

        if not only_one:
            message = self.message
            if message is None:
                message = field.gettext('It is required to set the value of one of the fields: %(fieldnames)s.')

            raise StopValidation(message % {'fieldnames': ', '.join(fieldnames)})

        if only_one and not field.data:
            raise StopValidation


class Phone(object):
    def __init__(self, region=None, keep_raw_input=False, _check_region=True, message=None):
        self.region = region
        self.keep_raw_input = keep_raw_input
        self._check_region = _check_region
        self.message = message or 'Invalid phone number.'

    def __call__(self, form, field):
        try:
            numobj = phonenumbers.parse(
                field.data, region=self.region, keep_raw_input=self.keep_raw_input, _check_region=self._check_region
            )
            if not phonenumbers.is_valid_number(numobj):
                raise ValidationError(self.message)
        except phonenumbers.NumberParseException as err:
            raise ValidationError(err)


class Slug(object):
    slug_re = r'^[a-z0-9-_]+$'
    slug_unicode_re = r'^[-\w]+$'

    def __init__(self, allow_unicode=False, message=None):
        self.allow_unicode = allow_unicode
        self.message = message

    def __call__(self, form, field):
        pattern = self.slug_unicode_re if self.allow_unicode else self.slug_re
        if not re.match(pattern, field.data or '', re.I):
            raise ValidationError(self.message or field.gettext('Invalid input.'))



"""
The module contains functions for working with E-Mails.
"""

import typing as t

from flask import (
    render_template,
    render_template_string,
    Flask,
)
from flask_mail import (
    email_dispatched,
    Mail,
    Message,
)


__all__ = (
    'get_mail_instance',
    'render_html_message',
)


def get_mail_instance(app: t.Optional[Flask] = None) -> Mail:
    """Returns the configured instance of the Flask-Mail extension."""
    mail = Mail(app)
    email_dispatched.connect(email_dispatched_handler)
    return mail


def email_dispatched_handler(
    message: Message,
    app: Flask,
) -> None:
    """Logs messages to the console in debug mode."""
    if app.debug or app.config['MAIL_SUPPRESS_SEND']:
        app.logger.debug(message.as_string())


def make_template(
    base: str,
    blocks: t.Dict[str, str],
) -> str:
    """Creates a Jinja template on the fly."""
    result = ['{%- extends "', base, '" -%}']

    for name, body in blocks.items():
        result.extend(['{% block ', name, ' %}', body, '{% endblock %}'])

    return ''.join(result)


def render_html_message(
    template: str,
    blocks: t.Optional[t.Dict[str, str]] = None,
    /,
    **context: t.Any,
) -> str:
    """
    Returns the body of the email in HTML format.

    Arguments:
        template (str): The name of the template to render.
        blocks (dict): The content of the blocks to be replaced.
        context (dict): The variables to make available in the template.
    """
    if blocks is None:
        return render_template(template, **context)
    _template = make_template(template, blocks)
    return render_template_string(_template, **context)



from flask import Blueprint


bp = Blueprint('users', __name__, url_prefix='/users')


@bp.get('/')
def index():
    return [
        {'id': 1, 'name': 'Linus Torvalds'},
        {'id': 2, 'name': 'Richard Stallman'},
    ]



from functools import wraps
import typing as t

from .utils import camel_to_snake


_F = t.TypeVar('_F', bound=t.Callable[..., t.Any])


__all__ = ('getattr_or_implement', 'route')


def getattr_or_implement(getter: _F) -> _F:
    """
    Used to get the value of a property in a getter method
    when the property is not known to exist or not.

    If the property does not exist, the getter method must be implemented.

    The getter method never takes any arguments.
    """
    @wraps(getter)
    def wrapper(self: t.Any) -> t.Any:
        try:
            return getter(self)
        except AttributeError as err:
            raise NotImplementedError(
                f'{err} - override the `{self.__class__.__name__}.{getter.__name__}()` method.'
            )
    return t.cast(_F, wrapper)


def setattr_or_implement(setter: _F) -> _F:
    """
    Used to set the value of a property in a setter method
    when the property is not known to exist or not.

    If the property does not exist, the setter method must be implemented.

    The setter method takes only one argument.
    """
    @wraps(setter)
    def wrapper(self: t.Any, value: t.Any) -> None:
        try:
            setter(self, value)
        except AttributeError as err:
            raise NotImplementedError(
                f'{err} - override the `{self.__class__.__name__}.{setter.__name__}()` method.'
            )
    return t.cast(_F, wrapper)


def route(obj, rule, *args, **kwargs):
    """Decorator for the View classes."""
    def decorator(cls):
        endpoint = kwargs.get('endpoint', camel_to_snake(cls.__name__))
        kwargs['view_func'] = cls.as_view(endpoint)
        obj.add_url_rule(rule, *args, **kwargs)
        return cls
    return decorator



from markupsafe import escape, Markup


class CheckboxGroup(object):
    """
    A widget for a group of checkboxes or radio buttons, depending on the type of elements.

    Attributes:
        inline (bool): Widgets are arranged in a row, not a list.
    """

    def __init__(self, inline=False):
        self.inline = inline

    def __call__(self, field, **kwargs):
        html = []
        css_class = 'form-check form-check-inline' if self.inline else 'form-check'

        for f in field:
            html.extend((
                f'<div class="{css_class}">',
                f(class_='form-check-input'),
                f.label(class_='form-check-label'),
                '</div>'
            ))

        return Markup(''.join(html))



from __future__ import annotations
import typing as t


__all__ = ('trim',)


def trim(
    s: t.Optional[str],
    chars: t.Optional[str] = None,
) -> t.Optional[str]:
    """
    To remove other characters when used as a filter for form fields, you can use `functools.partial`.

    Example:
        from functools import partial

        StringField(filters=[partial(trim, chars='/')])
        StringField(filters=[lambda s: trim(s, chars='/')])
    """
    return s.strip(chars) if isinstance(s, str) else None



from flask import Blueprint, url_for
from flask_useful import register_blueprints


bp = Blueprint('v2', __name__, url_prefix='/v2')
register_blueprints(bp, '.')


@bp.get('/')
def index():
    return {
        'users': url_for('.users.index', _external=True),
    }



from __future__ import annotations
import typing as t

from flask import views
from flask import request, url_for, redirect, render_template
from flask.typing import ResponseReturnValue
from flask_wtf import FlaskForm

from .utils import camel_to_list, make_url


class FormMixin(object):
    """
    Mixin to work with HTML-forms.

    Attributes:
        form_class (:py:class:`~flask_wtf.FlaskForm`):
            A reference to the base class of the form.
        success_endpoint (:obj:`str`):
            The endpoint to which to go if the form was processed successfully.
        success_endpoint_map (:obj:`dict`):
            Mapping the action name to an endpoint to navigate to if the form was successfully processed.
            The action name is the value of the value attribute of the submit button, or is specified manually.
    """

    form_class = None
    success_endpoint = None
    success_endpoint_map = None

    def get_form(self, *args: t.Any, **kwargs: t.Any) -> FlaskForm:
        """
        Returns:
            :py:class:`~flask_wtf.FlaskForm`: An instance of the form class.
        """
        return self.get_form_class()(*args, **kwargs)

    def get_form_class(self) -> t.Type[FlaskForm]:
        """
        Returns:
             :py:class:`~flask_wtf.FlaskForm`: A reference to the base class of the form.
        """
        if self.form_class is None:
            raise AttributeError(
                'You must assign the value of the attribute `form_class`, '
                f'or override the `{self.__class__.__name__}.get_form_class()` method.'
            )
        return self.form_class

    def form_valid(self, form):
        """Runs if the form is processed successfully."""
        raise NotImplementedError

    def form_invalid(self, form):
        """Runs if errors occurred while processing the form."""
        raise NotImplementedError

    def get_success_endpoint(self):
        """Returns the endpoint to which to go if the form was processed successfully."""
        if self.success_endpoint_map is not None:
            action_name = request.form.get('_action')
            return self.success_endpoint_map[action_name]

        if self.success_endpoint is not None:
            return self.success_endpoint

        raise AttributeError(
            'You must assign the value of the attribute `success_endpoint` or `success_endpoint_map`, '
            f'or override the `{self.__class__.__name__}.get_success_endpoint()` method.'
        )

    def make_redirect(self, obj=None):
        """
        Args:
            obj (object|dict): The object whose property values are used to build the URL.

        Returns:
            :class:`werkzeug.wrappers.Response`: The redirect to follow if the form was successfully processed.
        """
        endpoint = self.get_success_endpoint()
        success_url = url_for(endpoint) if obj is None else make_url(endpoint, obj)
        return redirect(success_url)

    def process_form(self):
        """
        Creates a form, validates and calls a method for further processing.

        Returns:
            :class:`werkzeug.wrappers.Response`: response object.
        """
        form = self.get_form()

        if form.validate_on_submit():
            return self.form_valid(form)

        return self.form_invalid(form)


class MethodView(views.MethodView):
    """
    Arguments:
        template_name (:obj:`str`):
            The name of the template, can be passed to the :py:meth:`~flask.views.View.as_view` method.

    Attributes:
        template_name (:obj:`str`): The name of the template.
    """

    template_name = None

    def __init__(self, template_name=None):
        if template_name:
            self.template_name = template_name

    def get_template_name(self) -> t.Union[str, t.List[str]]:
        """
        Returns the name of the template.

        If the template_name property is not set, the value will be generated automatically based on the class name.

        Example:
            >>> class MyEntityAction(MethodView): pass
            >>> view = MyEntityAction()
            >>> view.get_template_name()
            "my_entity/action.html"

        """
        if self.template_name is None:
            name = camel_to_list(self.__class__.__name__, lower=True)
            self.template_name = '{1}/{0}.html'.format(name.pop(), '_'.join(name))
        return self.template_name

    def render_template(self, **context: t.Any) -> str:
        """Render a template with passed context."""
        return render_template(self.get_template_name(), **context)


class FormView(MethodView, FormMixin):
    def get_object(self, *args: t.Any, **kwargs: t.Any) -> t.Optional[t.Any]:
        return None

    def get(self, *args: t.Any, **kwargs: t.Any) -> ResponseReturnValue:
        obj = self.get_object(*args, **kwargs)
        return self.render_template(obj=obj, form=self.get_form(obj=obj))

    def post(self, *args: t.Any, **kwargs: t.Any) -> ResponseReturnValue:
        obj = self.get_object(*args, **kwargs)
        form = self.get_form(obj=obj)

        if form.validate_on_submit():
            return self.form_valid(form, obj)

        return self.form_invalid(form, obj)

    def form_invalid(self, form: FlaskForm, obj: t.Any = None) -> ResponseReturnValue:
        """Runs if errors occurred while processing the form."""
        return self.render_template(obj=obj, form=form)

    def form_valid(self, form: FlaskForm, obj: t.Any = None) -> ResponseReturnValue:
        """Runs if the form is processed successfully."""
        raise NotImplementedError



from flask import Blueprint, url_for
from flask_useful import register_blueprints


bp = Blueprint('api', __name__, url_prefix='/api')
register_blueprints(bp, '.', include_packages=True)


@bp.route('/')
def index():
    return {
        'v1': url_for('.v1.index', _external=True),
        'v2': url_for('.v2.index', _external=True),
    }
